'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 6, 2016
@author: Niels Lubbes

The method "get_implicit_image()" should be called from 
"LinearSeries.get_implicit_image()".

The method "get_implicit_projection()" should be called from 
"LinearSeries.get_implicit_projection()".
'''

import warnings

from class_ls_tools import LSTools
from class_poly_ring import PolyRing
from get_linear_series import get_mon_lst

from sage_interface import sage__eval
from sage_interface import sage_PolynomialRing
from sage_interface import sage_QQ
from sage_interface import sage_SR
from sage_interface import sage_expand
from sage_interface import sage_Compositions
from sage_interface import sage_solve


def get_implicit_image( ls ):
    '''
    INPUT:
        - "ls" -- LinearSeries object. Elements in "ls.pol_lst"
                  should be homogeneous polynomials over QQ. 
                  These polynomials represent a map F between 
                  projective spaces. We assume that the polynomials 
                  are co-prime.
    OUTPUT
        - A list of polynomials in QQ[x0,...,xn]. 
          These polynomials represent the ideal of the image
          of the map F in projective n-space, where n is 
          "len(ls.pol_lst)-1".
          
          
          This method might not terminate within reasonable time.
    '''

    # QQ[x0,...,xn]
    vx_lst = [ 'x' + str( i ) for i in range( len( ls.pol_lst ) )]
    ring = sage_PolynomialRing( sage_QQ, vx_lst + list( ls.ring.gens() ) )
    x_lst = ring.gens()[0:len( vx_lst )]
    v_lst = ring.gens()[len( vx_lst ):]

    # coerce "ls.pol_lst""
    p_lst = [ sage__eval( str( pol ), ring.gens_dict() ) for pol in ls.pol_lst ]

    # construct ideal
    s_lst = [ x_lst[i] - p_lst[i] for i in range( len( p_lst ) )]
    s_ideal = ring.ideal( s_lst )
    LSTools.p( len( s_lst ), s_lst )

    # eliminate all variables except for the xi's
    e_lst = list( s_ideal.elimination_ideal( v_lst ).gens() )
    LSTools.p( len( e_lst ), e_lst )

    # test
    dct = { x_lst[i]:p_lst[i] for i in range( len( p_lst ) ) }
    r_lst = [ e.subs( dct ) for e in e_lst ]
    LSTools.p( 'test:', sum( r_lst ) == 0 )

    return e_lst


def get_implicit_projection( ls, deg ):
    '''
    This function does not work properly, since the output polynomial still contains
    undeterminate variables. These need to be solved.
    
    INPUT:
        - "ls"  -- LinearSeries s.t. "ls.pol_lst" consist of polynomials in x,y,z 
                   and of the same degree. Note that the polynomials in "ls.pol_lst"
                   define a birational map from the projective plane to a surface S in
                   projective n-space where n equals "len(ls.pol_lst)-1".
        - "deg" -- An integer representing the degree of the parametrized surface S.
    OUTPUT:
        - A polynomial F(x0:x1:x2:x3) of degree at most "deg" and 
          undetermined coefficients in (r0,r1,...). 
          The ".parent()" of the polynomial is "SymbolicRing".
          The zero-set V(F) is a projection of the surface S by the map
          (x0:x1:...:xn) |-> (x0:x1:x2:x3) 
    '''
    p0, p1, p2, p3 = ls.pol_lst[0:4]

    # construct a polynomial ring
    c_len = len( sage_Compositions( deg + 4, length = 4 ).list() )
    c_str_lst = [ 'c' + str( i ) for i in range( c_len )]

    R = sage_PolynomialRing( PolyRing.num_field, ['x0', 'x1', 'x2', 'x3'] + c_str_lst + ['x', 'y', 'z'], order = 'lex' )
    x0, x1, x2, x3 = R.gens()[0:4]
    c_lst = R.gens()[4:4 + c_len]
    x, y, z = R.gens()[4 + c_len:]
    m_lst = []
    for a, b, c, d  in sage_Compositions( deg + 4, length = 4 ):
        m_lst += [ x0 ** ( a - 1 ) * x1 ** ( b - 1 ) * x2 ** ( c - 1 ) * x3 ** ( d - 1 ) ]


    R_dict = R.gens_dict()
    R_dict.update( PolyRing.num_field.gens_dict() )

    p0, p1, p2, p3 = sage__eval( str( [p0, p1, p2, p3] ), R_dict )

    LSTools.p( R )
    LSTools.p( 'm_lst =', m_lst )
    LSTools.p( '(p0, p1, p2, p3) =', ( p0, p1, p2, p3 ) )

    # construct polynomial in x0, x1, x2 with coefficients in c_lst
    F = 0
    for i in range( len( m_lst ) ):
        F += c_lst[i] * m_lst[i]
    LSTools.p( 'F =', F )

    # compute F( p0(x,y,z):...: p3(x,y,z) )
    FP = sage_expand( F.subs( {x0:p0, x1:p1, x2:p2, x3:p3} ) )  # expand to prevent a bug in sage.
    LSTools.p( 'FP =', FP )

    # obtain coefficients w.r.t. x, y and z
    coef_lst = []
    pmzdeg = p0.total_degree()
    comp_lst = sage_Compositions( deg * pmzdeg + 3 , length = 3 ).list()  # e.g. [1,1,2]=[0,0,1]
    LSTools.p( 'comp_lst =', comp_lst )
    for comp in comp_lst:
        coef = FP.coefficient( {x: comp[0] - 1, y: comp[1] - 1, z:comp[2] - 1} )
        coef_lst += [coef]
    LSTools.p( 'coef_lst =', coef_lst )

    # compute groebner basis and reduce F w.r.t. the GB
    #
    # alternative code
    # ----------------
    # GB = list( R.ideal( coef_lst ).groebner_basis() )
    # LSTools.p( 'GB =', GB )
    # red_F = F.reduce( R.ideal( GB ) )
    # for g in GB:
    #    red_F = red_F.quo_rem( g )[1]
    # LSTools.p( 'red_F =', red_F )
    # ----------------
    #
    scoef_lst = [ sage_SR( coef ) for coef in coef_lst]
    sc_lst = [ sage_SR( c ) for c in c_lst]
    sol_lst = sage_solve( scoef_lst, sc_lst, solution_dict = True )
    LSTools.p( sol_lst )
    red_F = sage_expand( sage_SR( F ).subs( sol_lst[0] ) )
    LSTools.p( 'red_F =', red_F )

    # check the solutions
    sp0, sp1, sp2, sp3 = sage_SR( p0 ), sage_SR( p1 ), sage_SR( p2 ), sage_SR( p3 )
    sx0, sx1, sx2, sx3 = sage_SR( x0 ), sage_SR( x1 ), sage_SR( x2 ), sage_SR( x3 )
    chk_F = sage_expand( red_F.subs( {sx0:sp0, sx1:sp1, sx2:sp2, sx3:sp3} ) )
    LSTools.p( 'chk_F =', chk_F )
    if chk_F != 0:
        warnings.warn( 'The polynomial red_F(p0:p1:p2:p3) does not vanish.' )

    return red_F



'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 6, 2016
@author: Niels Lubbes

The method "get_linear_series()" should be called from "LinearSeries.get()".
'''

from class_ls_tools import LSTools
from class_poly_ring import PolyRing
import class_linear_series

from sage_interface import sage__eval
from sage_interface import sage_Compositions
from sage_interface import sage_matrix
from sage_interface import sage_vector


def get_ls_lst( ls, bp_lst ):
    '''
    Helper for "get_linear_series" method. 
    '''
    ind_str = None

    ls_lst = []
    for bp in bp_lst:

        if ind_str == None:
            ind_str = bp.depth * '\t'
            LSTools.p( ind_str + 'input  = len( bp_lst ):' + str( len( bp_lst ) ) + ',' + str( ls ) )

        # base points which do not have multiplicity at least
        # can be ignored
        #
        if bp.mult <= 0: continue
        LSTools.p( 'basepoint  = ' + str( bp )[1:].replace( '\n', '\n' + ind_str ) )

        # translate the base point to the origin
        #
        nls = ls.copy().translate_to_origin( bp.sol )
        LSTools.p( ind_str + 'translated =', nls )

        #
        # Loop through (a,b) s.t.
        # (a-1)+(b-1)=bp.mult-1 and (a-1)>=(b-1)>=0
        #
        for m in range( 1, bp.mult + 1 ):
            for a, b in sage_Compositions( m - 1 + 2, length = 2 ):
                ls_lst += [ nls.copy().diff( a - 1, b - 1 ) ]
                LSTools.p( ind_str + str( a - 1 ) + ',' + str( b - 1 ) + ',' + str( ls_lst[-1] ) )

        #
        # Continue recursively.
        #
        u, v = nls.gens()
        ls_lst += get_ls_lst( nls.copy().subs( {u:u * v} ).quo( v ** bp.mult ), bp.bp_lst_t )
        ls_lst += get_ls_lst( nls.copy().subs( {v:v * u} ).quo( u ** bp.mult ), bp.bp_lst_s )

    if ind_str != None:
        LSTools.p( ind_str + 'output =', [str( ls ) for ls in ls_lst] )

    return ls_lst


def get_mon_lst( deg_lst, g_lst ):
    '''
    INPUT:
        - "deg_lst" -- A list of either one or two integers representing the degree of polynomials.
        - "g_lst"   -- List of either 3 or 4 generators of a ring. 
    OUTPUT:
        - If "gens" consist of 3 generators then returns
          a list of all monomials in "gens" of degree "deg_tup[0]".
        - If "gens" consist of 4 generators then returns
          a list of all monomials in "gens" of bi-degree ("deg_tup[0]","deg_tup[1]").
    EXAMPLE:
        - get_mon_lst( [2], PolyRing('x,y,z').gens() ) == [ z^2, y*z, x*z, y^2, x*y, x^2 ]
        
        - get_mon_lst( [2,2], PolyRing('x,y,v,w').gens() ) == [ x^2*v^2, x^2*v*w, x^2*w^2, 
                                                                x*y*v^2, x*y*v*w, x*y*w^2, 
                                                                y^2*v^2, y^2*v*w, y^2*w^2 ] 
    '''

    mon_lst = []
    if len( g_lst ) == 4:
        x, y, v, w = g_lst
        for a, b, c, d in sage_Compositions( sum( deg_lst ) + 4, length = 4 ):
            if a + b == deg_lst[0] + 2 and c + d == deg_lst[1] + 2:
                mon_lst += [ x ** ( a - 1 ) * y ** ( b - 1 ) * v ** ( c - 1 ) * w ** ( d - 1 ) ]
    elif len( g_lst ) == 3:
        x, y, z = g_lst
        for a, b, c in sage_Compositions( deg_lst[0] + 3, length = 3 ):
            mon_lst += [ x ** ( a - 1 ) * y ** ( b - 1 ) * z ** ( c - 1 ) ]
    else:
        raise ValueError( 'Expect either 3 or 4 generators:', g_lst )

    return mon_lst


def get_linear_series( deg_lst, bp_tree ):
    '''
    INPUT:
        - "deg_lst" -- A list of either one or two integers representing the degree of polynomials.
        - "bp_tree" -- BasePointTree where base points might be in 
                       overlapping charts.
                       We require that "bp_tree.chart_lst" equals either 
                       ['z', 'x', 'y'] or ['xv', 'xw', 'yv', 'yw'].                       
    OUTPUT:
        - Return A LinearSeries "ls" with base points defined by "bp_tree".
          The polynomials in "ls.pol_lst" are either 
                  * homogeneous in (x:y:z) and of degree "deg_tup[0]" or          
                  * bi-homogenous in (x:y)(v:w) and of bi-degree ("deg_tup[0]","deg_tup[1]")
          Note that there might unassigned base points. 
    '''

    #
    # Obtain generators of polynomial ring from "bp_tree.chart_lst"
    # Note that for initializing PolyRing, it is important that the
    # base field is not reset to the rationals QQ. The input "bp_tree"
    # could already be defined over some extension field.
    #
    if bp_tree.chart_lst == ['z', 'x', 'y']:
        ring = PolyRing( 'x,y,z', False )
    elif bp_tree.chart_lst == ['xv', 'xw', 'yv', 'yw']:
        ring = PolyRing( 'x,y,v,w', False )
    else:
        raise ValueError( 'Expect "bp_tree.chart_lst" in ', ( ['z', 'x', 'y'], ['xv', 'xw', 'yv', 'yw'] ) )

    #
    # Obtain linear series defined by
    # monomials of degree "deg_lst[0]" or bidegree ("deg_lst[0]","deg_lst[1]").
    #
    mon_lst = get_mon_lst( deg_lst, ring.gens() )
    LSTools.p( 'mon_lst =', mon_lst )
    ls = class_linear_series.LinearSeries( mon_lst, ring )

    #
    # For each chart "c" we obtain a list of linear series.
    #
    ls_lst = []
    for c in bp_tree.chart_lst:
        ls_lst += get_ls_lst( ls.copy().chart( c ), bp_tree[c] )

    #
    # Create matrix whose coefficients are obtained
    # by evaluating each polynomial in linear series
    # in "ls_lst" at (0,0).
    #
    row_lst = []
    for ls in ls_lst:
        LSTools.p( ls )
        for g in ls.ring.gens():
            ls.subs( {g:0} )
        row = sage__eval( str( ls.pol_lst ), PolyRing.num_field.gens_dict() )
        row_lst += [row]
    LSTools.p( 'matrix =', row_lst )

    #
    # Compute kernel.
    #
    kern = list( sage_matrix( row_lst ).right_kernel().matrix() )
    LSTools.p( 'kernel =', kern )
    LSTools.p( mon_lst )

    #
    # Obtain linear series from linear conditions on "mon_lst".
    #
    if kern != []:
        mon_lst = ring.coerce( mon_lst )  # update w.r.t. PolyRing.num_field
        kern = ring.coerce( kern )
        pol_lst = sage_matrix( kern ) * sage_vector( mon_lst )
    else:
        pol_lst = []

    LSTools.p( pol_lst )

    return class_linear_series.LinearSeries( pol_lst, ring )





'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 6, 2016
@author: Niels Lubbes

The method "get_solution_set()" should be called 
from "LinearSeries.get_solution_set()".
'''

from class_ls_tools import LSTools

from sage_interface import sage_Combinations
from sage_interface import sage_gcd


def get_solution_set( ls ):
    '''
    INPUT:
        - "ls" -- Assumes that "ls.pol_lst" is a list of polynomials 
                  in two variables and that its zero-set is 0-dimensional.
        
    OUTPUT:
        - The number field of "ls.ring" is extended with roots in the zero-set. 
        - Returns zero-set of "ls.pol_lst". Thus the set of points on which 
          all polynomials in "ls.pol_lst" vanish.
    '''

    LSTools.p( 'input =', ls )

    if  len( ls.gens() ) != 2:
        raise Exception( 'Not implemented for polynomials in', ls.gens() )

    # Try all possible pairs of polynomials in pol_lst
    # until their resultant wrt. y is nonzero.
    xres = 0
    x, y = ls.gens()
    interval = range( len( ls.pol_lst ) )
    for i, j in sage_Combinations( interval, 2 ):
        xres = ls.ring.resultant( ls.pol_lst[i], ls.pol_lst[j], y )
        if xres != 0:
            break
    if xres == 0:
        if sage_gcd( ls.pol_lst ) != 1:
            raise ValueError( 'Expect gcd of "pol_lst" to be 1:', sage_gcd( ls.pol_lst ) )
        # TODO: handle this case as well.
        raise Exception( 'All pairs of polynomials in "pol_lst" have a common factor: ', ls.pol_lst )

    # extend number field with roots of resultant
    tres = str( xres ).replace( str( x ), 't' )
    ls.ring.ext_num_field( tres )
    ls.pol_lst = ls.ring.coerce( ls.pol_lst )
    xres = ls.ring.coerce( xres )
    x, y = ls.ring.gens()

    # factor resultant in linear factors
    fct_lst = ls.ring.factor( xres )

    # obtain candidates for x-coordinates of solutions
    xsol_lst = []
    for fct in fct_lst:
        xsol_lst += [ -fct[0].subs( {xres.variables()[0]:0} ) ]
    xsol_lst = list( set( xsol_lst ) )
    LSTools.p( xsol_lst )

    # find y-coordinate for each x-coordinate
    xysol_lst = []
    for xsol in xsol_lst:

        # substitute x-coord
        xsol = ls.ring.coerce( xsol )  # ring.num_field is changed during while loop
        ypol_lst = [ pol.subs( {x:xsol} ) for pol in ls.pol_lst]
        ypol_lst = [ ypol for ypol in ypol_lst if ypol != 0 ]
        ypol = ypol_lst[0]

        # extend number field with roots of polynomial in y
        tpol = str( ypol ).replace( str( y ), 't' )
        ls.ring.ext_num_field( tpol )
        ls.pol_lst = ls.ring.coerce( ls.pol_lst )
        ypol = ls.ring.coerce( ypol )
        ypol_lst = ls.ring.coerce( ypol_lst )
        x, y = ls.ring.gens()

        # obtain candidates for y-coordinates
        ysol_lst = []
        if ypol not in ls.ring.get_num_field():
            for fct in ls.ring.factor( ypol ):
                ysol_lst += [ -fct[0].subs( {ypol.variables()[0]:0} ) ]

        # check whether (xsol, ysol) is in the zero-set of "ls.pol_lst"
        for ysol in ysol_lst:
            lst = [ ypol.subs( {y:ysol} ) for ypol in ypol_lst ]
            if set( lst ) == {0}:
                xysol_lst += [( xsol, ysol )]

    xysol_lst = list( set( xysol_lst ) )
    LSTools.p( 'output =', xysol_lst )

    return xysol_lst



'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 4, 2016
@author: Niels Lubbes
'''

from class_poly_ring import PolyRing
from get_base_point_tree import get_bp_tree
from get_solution_set import get_solution_set
from get_linear_series import get_linear_series
from get_implicit import get_implicit_projection
from get_implicit import get_implicit_image


class LinearSeries:
    '''
    A linear series in the plane is represented 
    as a list of polynomials in "PolyRing". 
    '''

    def __init__( self, pol_lst = [], ring = PolyRing() ):
        '''
        INPUT:
            - "pol_lst" -- A list of (strings of) polynomials in "ring".
            - "ring"    -- A "PolyRing" representing a polynomial ring.   
        OUTPUT:
            - "self.pol_lst"
            - "self.ring"
        '''
        self.ring = ring
        self.pol_lst = [self.ring.coerce( pol ) for pol in pol_lst]


    def get_bp_tree( self ):
        return get_bp_tree( self )


    def get_solution_set( self ):
        return get_solution_set( self )


    @staticmethod
    def get( deg_lst, bp_tree ):
        return get_linear_series( deg_lst, bp_tree )


    def get_implicit_projection( self, deg ):
        return get_implicit_projection( self, deg )


    def get_implicit_image( self ):
        return get_implicit_image( self )


    def copy( self ):
        cp_ring = PolyRing( self.gens() )
        cp_pol_lst = cp_ring.coerce( self.pol_lst )
        return LinearSeries( cp_pol_lst, cp_ring )


    def gens( self ):
        return self.ring.gens()


    def subs( self, sub_dct ):
        '''
        INPUT:
            - "self"     -- "self.ring" is a PolyRing and "self.polyring" is a list
                            of polynomials in PolyRing.
            - "sub_dct"  --  A dictionary with [key] and [value] elements in PolyRing.
        OUTPUT:
            - Sets "self.pol_lst" to the list of polynomials in PolyRing where
              each occurrence of [key] is substituted with [value].   
        '''
        self.pol_lst = [ self.ring.coerce( pol ) for pol in self.pol_lst ]
        self.pol_lst = [ pol.subs( sub_dct ) for pol in self.pol_lst ]
        return self


    def chart( self, c_lst ):
        '''        
        INPUT:
            - "self"  -- The polynomials in "self.pol_lst" are assumed 
                         to be (bi-) homogeneous in PolyRing.
            - "c_lst" -- A list of characters where each character  
                         denotes a generator of the polynomial ring "self.ring" 
                         of type PolyRing. For bi-homogeneous polynomials, 
                         we expect "c_lst" to be a string of length two,
                         since we can consider a string as a list of characters.                                                    
        OUTPUT:
            - Substitutes {c:1} in polynomials in "self.pol_lst" 
              for all c in "c_lst". 
              The generators of the polynomial ring "self.ring" 
              are modified accordingly. 
              Thus "self" now represents a linear series 
              in the affine plane.
            
            - Returns "self". 
        '''
        for c in c_lst:
            var = self.ring.coerce( c )
            self.subs( { var:1} )
            new_gens = [ gen for gen in self.gens() if gen != var ]
            self.ring.set_gens( new_gens )
            self.pol_lst = self.ring.coerce( self.pol_lst )

        return self


    def diff( self, mu, mv ):
        '''
        INPUT:
            - "self" -- "self.pol_lst" are bivariate polynomials in PolyRing in say u and v.
            - "mu"   -- An integer. 
            - "mv"   -- An integer.
        OUTPUT:
            - "self.pol_lst" is set to
             
                [ d_u^{mu}d_v^{mv}(f) for f in "pol_lst"  ] 
              
              where d_u(.)=diff(.,u) denotes differentiation w.r.t. u.
              
              Thus we consider the (multiple) partial derivatives of each polynomial 
              in "self.pol_lst" wrt the variable u and v.
            
            - Returns "self".
        '''
        u, v = self.ring.gens()
        self.pol_lst = [ self.ring.diff( pol, u, mu ) for pol in self.pol_lst ]
        self.pol_lst = [ self.ring.diff( pol, v, mv ) for pol in self.pol_lst ]
        return self


    def translate_to_origin( self, sol ):
        '''
        INPUT:
            - "sol" -- A 2-tuple of elements in "self.ring.get_num_field()".
            
        METHOD:
            - Composes each polynomial F in "self.pol_lst" with a
              translation T such that F(p)=(FoT)(0,0) 
              where p=(sol[0],sol[1]).               
        '''
        u, v = self.gens()
        self.subs( {u:u + sol[0], v:v + sol[1]} )
        return self


    def blow_up_origin( self, chart ):
        '''
        INPUT:
            - "self"  -- Polynomials in PolyRing in the list "self.pol_lst" 
                         should be bi-variate and co-prime.
            - "chart" -- Either 's' or 't'.
            
        OUTPUT:
            - Returns the following 2 values:
                      * "self" 
                      * multiplicity of origin in "self.pol_lst".
        
            - Additionally "self" is modified as follows.   
            
              Let (g_i(u,v))_i denote the list of polynomials in "self.pol_lst" 
              indexed by i and in the variables u an v. These polynomials are elements
              in PolyRing.
              
              Let "mul" denote the multiplicity of the linear series at the origin: 
                  mul = deg( gcd( (g_i(u*v,v)_i) ) )
                           
              If "chart" equals 't' then "self.pol_lst" is modified to
                  ( g_i(u*v,v)*v^{-mul} )_i
              
              If "chart" equals 's' then "self.pol_lst" is modified to
                  ( g_i(u,v*u)*u^{-mul} )_i

              The linear series represented by "self.pol_lst" is after this
              modification isomorphic to a chart of the blow up of a linear series 
              at the origin. 
              
              The mathematical reason is as follows: 
              
              Let g(u,v) = c0*g0(u,v)+...+c5*g5(u,v) for some general coefficients ci in a 
              number field.
              
              In the chart z!=0, the blow up at the origin (0,0) is in coordinates:
                  
                U_z = { (u,v)x(s:t) | u*t=v*s and g(u,v)=0 } in A^2xP^1,
              
              and thus contained in the fiber product of the affine plane 
              with the projective line.  
                
              If also t!=0 then we obtain the chart
              
                U_t = { (u,v)x(s/t) | u=v*(s/t) and g(u,v)=0  } in A^2xA^1.
              
              The chart U_t has a component coming from the 
              multiplicity of g(u,v) at the origin (this is known
              in the algebraic geometry literature as 
              the exceptional component). This multiplicity is equal 
              to "mul" as above.
              
              The chart U_t is isomorphic to 
                  
                U'_t = { (s/t, v) | g( v*(s/t), v)=0 } in A^2.
              
              where u is recovered from (s/t, v) since u=v*(s/t).
              So we can subsequently repeat this blow up procedure,
              by setting the new u as u:=s/t, and factoring out
              the exceptional component.
        '''
        u, v = self.gens()
        if chart == 't':
            self.subs( {u:u * v} )
            f = v
        elif chart == 's':
            self.subs( {v:v * u} )
            f = u
        else:
            raise ValueError( 'Chart should be either "t" or "s"', chart )

        gcd_fct_lst, self.pol_lst = self.ring.aux_gcd( self.pol_lst )

        if len( gcd_fct_lst ) > 1:
            raise Exception( 'Linear series has a fixed component.' )

        for fct in gcd_fct_lst:
            if fct[0] != f:
                raise Exception( 'Expect factor: ' + str( f ) + '^mult.', gcd_fct_lst )
            return self, fct[1]

        return self, 0


    def quo( self, pol2 ):
        '''
        INPUT:
            - "pol2" -- A (string of a) polynomial "pol" in PolyRing "self.ring".
             
        OUTPUT:
            - Polynomials in "self.pol_lst" are replaced with their quotient with "pol2".            
            - Return "self"
        '''
        pol2 = self.ring.coerce( pol2 )
        self.pol_lst = [ self.ring.quo( pol1, pol2 ) for pol1 in self.pol_lst ]

        return self


    def __str__( self ):
        '''                
        OUTPUT:
          - Human readable string representation of object. The String
            has the following format:
                { N, <<F0,...,FN>>, R }
            where
                * N         = Number of polynomials defining the linear series.
                * F0,...,FN = The N polynomials that define the linear series.
                * R         = The PolyRing over which the linear series is defined 
                
        EXAMPLES:
          - sage: print( LinearSeries( ['x^2', 'x*z + y^2'], PolyRing( 'x,y,z', True ) ) )                          
            out : { 2, <<x^2, y^2 + x*z>>, QQ[x, y, z] }
              
          - sage: ring = PolyRing( 'x,y,z', True )
            sage: ring.ext_num_field( 't^2 + 1' )              
            sage: print( LinearSeries( ['x^2+a0*y*z','y^2+x*z'], ring ) )
            out : { 2, <<x^2 + (a0)*y*z, y^2 + x*z>>, QQ( <a0|t^2 + 1> )[x, y, z] }                   
        '''

        s = '{ '
        s += str( len( self.pol_lst ) ) + ', '
        s += '<<' + str( self.pol_lst )[1:-1] + '>>, '
        s += str( self.ring )
        s += ' }'
        return s





'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.

Created on Aug 6, 2016
@author: Niels Lubbes
'''

from class_ls_tools import LSTools
from class_poly_ring import PolyRing
from class_linear_series import LinearSeries
from class_base_points import BasePointTree

from sage_interface import sage_ZZ
from sage_interface import sage_QQ
from sage_interface import sage_invariant_theory
from sage_interface import sage__eval
from sage_interface import sage_PolynomialRing
from sage_interface import sage_matrix
from sage_interface import sage_diagonal_matrix
from sage_interface import sage_vector
from sage_interface import sage_SR
from sage_interface import sage_var
from sage_interface import sage_diff
from sage_interface import sage_factor


def usecase__get_base_points__P2():
    '''
    We obtain (infinitely near) base points of a 
    linear series defined over a number field.
    '''
    ring = PolyRing( 'x,y,z', True )
    ring.ext_num_field( 't^2 + 1' )
    ring.ext_num_field( 't^3 + a0' )

    a0, a1 = ring.root_gens()
    x, y, z = ring.gens()

    pol_lst = [x ** 2 + a0 * y * z, y + a1 * z + x ]

    ls = LinearSeries( pol_lst, ring )
    LSTools.p( ls )

    bp_tree = ls.get_bp_tree()
    LSTools.p( bp_tree )


def usecase__get_base_points__P1P1():
    '''
    We obtain (infinitely near) base points of a 
    linear series defined by bi-homogeneous polynomials.
    Such polynomials are defined on P^1xP^1 (the fiber product 
    of the projective line with itself). 
    The coordinate functions of P^1xP^1 are (x:y)(v:w).
    '''

    pol_lst = [ '17/5*v^2*x^2 + 6/5*v*w*x^2 + w^2*x^2 + 17/5*v^2*y^2 + 6/5*v*w*y^2 + w^2*y^2',
                '6/5*v^2*x^2 + 8/5*v*w*x^2 + 6/5*v^2*y^2 + 8/5*v*w*y^2',
                '7/5*v^2*x^2 + 6/5*v*w*x^2 + w^2*x^2 + 7/5*v^2*y^2 + 6/5*v*w*y^2 + w^2*y^2',
                '4*v^2*x*y',
                '-2*v^2*x^2 + 2*v^2*y^2',
                '2/5*v^2*x^2 + 6/5*v*w*x^2 - 4*v^2*x*y - 2/5*v^2*y^2 - 6/5*v*w*y^2',
                '2*v^2*x^2 + 4/5*v^2*x*y + 12/5*v*w*x*y - 2*v^2*y^2' ]

    ls = LinearSeries( pol_lst, PolyRing( 'x,y,v,w' ) )
    LSTools.p( ls )

    bp_tree = ls.get_bp_tree()
    LSTools.p( bp_tree )


def usecase__get_base_points__examples():
    '''
    We obtain (infinitely near) base points of several 
    examples of linear series in order to test
    "get_base_point_tree.get_bp_tree()".
    '''
    pol_lst_lst = []

    # 0
    pol_lst_lst += [['x^2*z + y^2*z', 'y^3 + z^3']]

    # 1
    pol_lst_lst += [['x^3*z^2 + y^5', 'x^5']]

    # 2 (example from PhD thesis)
    pol_lst_lst += [['x^2', 'x*z + y^2']]

    # 3
    pol_lst_lst += [['x^2 + y^2', 'x*z + y^2']]

    # 4
    pol_lst_lst += [['x*z', 'y^2', 'y*z', 'z^2']]

    # 5
    pol_lst_lst += [['x^5', 'y^2*z^3']]

    # 6
    pol_lst_lst += [['x^2*y', 'x*y^2', 'x*y*z', '( x^2 + y^2 + z^2 )*z' ]]

    # 7 (weighted projective plane P(2:1:1))
    pol_lst_lst += [['z^6'    , 'y*z^5'    , 'y^2*z^4'    ,
                     'y^3*z^3', 'y^4*z^2'  , 'y^5*z'      , 'y^6',
                     'x^2*z^4', 'x^2*y*z^3', 'x^2*y^2*z^2', 'x^3*z^3']]

    # 8 (provided by Martin Weimann)
    pol_lst_lst += [['y^2*(x^3+x^2*z+2*x*z^2+z^3)+y*(2*x^3+2*x*z^2)*z+z^2*(x^3-x^2*z+x*z^2)',
                    'z^5' ]]

    # 9
    pol_lst_lst += [['x^3*y^4+x*y^4*z^2+3*x^2*y^3*z^2+3*x*y^2*z^4+y',
                    'z^7']]

    # 10 (degree 6 del Pezzo in S^5, which is the orbital product of 2 circles)
    pol_lst_lst += [[
        'x^2*y^2 + 6/5*x^2*y*z + 17/5*x^2*z^2 + y^2*z^2 + 6/5*y*z^3 + 17/5*z^4',
        '8/5*x^2*y*z + 6/5*x^2*z^2 + 8/5*y*z^3 + 6/5*z^4',
        'x^2*y^2 + 6/5*x^2*y*z + 7/5*x^2*z^2 + y^2*z^2 + 6/5*y*z^3 + 7/5*z^4',
        '4*x*z^3',
        '2*x^2*z^2 - 2*z^4',
        '-6/5*x^2*y*z - 2/5*x^2*z^2 - 4*x*z^3 + 6/5*y*z^3 + 2/5*z^4',
        '-2*x^2*z^2 + 12/5*x*y*z^2 + 4/5*x*z^3 + 2*z^4'
       ]]


    BasePointTree.short = True

    idx = 0
    for pol_lst in pol_lst_lst:

        ls = LinearSeries( pol_lst, PolyRing( 'x,y,z', True ) )

        LSTools.p( 3 * ( '\n' + 50 * '=' ) )
        LSTools.p( 'index for example =', idx )
        LSTools.p( ls )

        bp_tree = ls.get_bp_tree()
        LSTools.p( bp_tree )

        idx += 1


def usecase__get_linear_series__P2():
    '''
    Construct linear series of curves in the plane P^2, 
    with a given tree of (infinitely near) base points.
    '''

    # Example from PhD thesis (page 159).
    PolyRing.reset_base_field()
    bp_tree = BasePointTree()
    bp = bp_tree.add( 'z', ( 0, 0 ), 1 )
    bp = bp.add( 't', ( 0, 0 ), 1 )
    bp = bp.add( 't', ( -1, 0 ), 1 )
    bp = bp.add( 't', ( 0, 0 ), 1 )
    LSTools.p( bp_tree )

    ls = LinearSeries.get( [2], bp_tree )
    LSTools.p( ls.get_bp_tree() )


def usecase__get_linear_series__P1P1_DP6():
    '''
    Construct linear series of curves in P^1xP^1, 
    with a given tree of (infinitely near) base points.
    '''

    # construct ring over Gaussian rationals
    #
    ring = PolyRing( 'x,y,v,w', True )
    ring.ext_num_field( 't^2 + 1' )
    a0 = ring.root_gens()[0]

    # setup base point tree for 2 simple complex conjugate base points.
    #
    bp_tree = BasePointTree( ['xv', 'xw', 'yv', 'yw'] )
    bp = bp_tree.add( 'xv', ( -a0, a0 ), 1 )
    bp = bp_tree.add( 'xv', ( a0, -a0 ), 1 )
    LSTools.p( 'We consider linear series of curves in P^1xP^1 with the following base point tree:' )
    LSTools.p( bp_tree )

    #
    # Construct linear series is defined by a list of polynomials
    # in (x:y)(v:w) of bi-degree (2,2) with base points as in "bp_tree".
    # The defining polynomials of this linear series, correspond to a parametric map
    # of an anticanonical model of a Del Pezzo surface of degree 6 in P^6.
    #
    # We expect that the linear series is defined by the following set of polynomials:
    #
    # [ 'x^2*v^2 - y^2*w^2', 'x^2*v*w + y^2*v*w', 'x^2*w^2 + y^2*w^2', 'x*y*v^2 - y^2*v*w',
    #   'x*y*v*w - y^2*w^2', 'y^2*v*w + x*y*w^2', 'y^2*v^2 + y^2*w^2' ]
    #
    ls = LinearSeries.get( [2, 2], bp_tree )
    LSTools.p( 'The linear series of bi-degree (2,2) corresponding to this base point tree is as follows:' )
    LSTools.p( ls.get_bp_tree() )

    #
    # construct corresponding linear series of bi-degree (1,1)
    # and 2 simple complex conjugate base points
    #
    # We expect that the linear series is defined by the following set of polynomials:
    #
    # ['x * v - y * w', 'y * v + x * w' ]
    #
    #
    ls = LinearSeries.get( [1, 1], bp_tree )
    LSTools.p( 'The linear series of bi-degree (1,1) corresponding to this base point tree is as follows:' )
    LSTools.p( ls.get_bp_tree() )


def usecase__get_implicit__DP6():
    '''
    Construct linear series of curves in P^1xP^1, 
    with a given tree of (infinitely near) base points.
    The defining polynomials of this linear series, correspond to a parametric map
    of an anticanonical model of a Del Pezzo surface of degree 6 in P^6.
    Its ideal is generated by quadratic forms. We search for a quadratic form in
    this ideal of signature (1,6). This quadratic form corresponds to a hyperquadric 
    in P^6, such that the sextic Del Pezzo surface is contained in this hyperquadric.         
    We construct a real projective isomorphism from this hyperquadric to the projectivization
    of the unit 5-sphere in P^6. 
    '''

    #
    # parametrization of degree 6 del Pezzo surface in projective 6-space.
    # See ".usecase__get_linear_series__P1P1_DP6()" for the construction of
    # this linear series.
    #
    pmz_lst = [
         'x^2*v^2 - y^2*w^2',
         'x^2*v*w + y^2*v*w',
         'x^2*w^2 + y^2*w^2',
         'x*y*v^2 - y^2*v*w',
         'x*y*v*w - y^2*w^2',
         'y^2*v*w + x*y*w^2',
         'y^2*v^2 + y^2*w^2'
         ]
    ls = LinearSeries( pmz_lst, PolyRing( 'x,y,v,w', True ) )

    #
    # base points of linear series
    #
    bp_tree = ls.get_bp_tree()
    LSTools.p( 'parametrization    =', ls.pol_lst )
    LSTools.p( 'base points        =' + str( bp_tree ) )

    #
    # implicit image in projective 6-space
    # of map associated to linear series
    #
    imp_lst = ls.get_implicit_image()
    LSTools.p( 'implicit equations =', imp_lst )

    #
    # compute Hilbert polynomial in QQ[x0,...,x6]
    #
    ring = sage_PolynomialRing( sage_QQ, [ 'x' + str( i ) for i in range( 7 )] )
    x_lst = ring.gens()
    imp_lst = sage__eval( str( imp_lst ), ring.gens_dict() )
    hpol = ring.ideal( imp_lst ).hilbert_polynomial()
    hdeg = hpol.diff().diff()
    LSTools.p( 'Hilbert polynomial =', hpol )
    LSTools.p( 'implicit degree    =', hdeg )

    #
    # equation of unit sphere is not in the ideal
    #
    s_pol = sum( [-x_lst[0] ** 2] + [x ** 2 for x in x_lst[1:]] )
    LSTools.p( 'Inside sphere?: ', s_pol in ring.ideal( imp_lst ) )

    #
    # compute random quadrics containing del Pezzo surface
    # until a quadric with signature (1,6) is found
    # or set a precomputed quadric with given "c_lst"
    #
    LSTools.p( 'Look for quadric in ideal of signature (1,6)...(may take a while)...' )
    sig = []
    while sorted( sig ) != [0, 1, 6]:

        # set coefficient list
        c_lst = []
        c_lst = [-1, -1, 0, 0, 0, -1, 1, 0, -1, -1, -1]  # uncomment to speed up execution
        if c_lst == []:
            lst = [-1, 0, 1]
            for imp in imp_lst:
                idx = int( sage_ZZ.random_element( 0, len( lst ) ) )
                c_lst += [ lst[idx] ]

        # obtain quadric in ideal from "c_lst"
        i_lst = range( len( imp_lst ) )
        M_pol = [ c_lst[i] * imp_lst[i] for i in i_lst if imp_lst[i].total_degree() == 2 ]
        M_pol = sum( M_pol )

        # eigendecomposition
        M = sage_invariant_theory.quadratic_form( M_pol, x_lst ).as_QuadraticForm().matrix()
        M = sage_matrix( sage_QQ, M )
        vx = sage_vector( x_lst )
        D, V = M.eigenmatrix_right()

        # determine signature of quadric
        num_pos = len( [ d for d in D.diagonal() if d > 0 ] )
        num_neg = len( [ d for d in D.diagonal() if d < 0 ] )
        num_zer = len( [ d for d in D.diagonal() if d == 0 ] )
        sig = [ num_pos, num_neg, num_zer ]
        LSTools.p( '\t sig =', sig, c_lst )

    #
    # output of M.eigenmatrix_right() ensures that
    # D has signature either (--- --- +) or (- +++ +++)
    #
    if num_pos < num_neg:  # (--- --- +) ---> (+ --- ---)
        V.swap_columns( 0, 6 )
        D.swap_columns( 0, 6 )
        D.swap_rows( 0, 6 )

    #
    # diagonal orthonormalization
    # note: M == W.T*D*W == W.T*L.T*J*L*W = U.T*J*U
    #
    W = sage_matrix( [col / col.norm() for col in V.columns()] )
    J = []
    for d in D.diagonal():
        if d > 0:
            J += [1]
        elif d < 0:
            J += [-1]
    J = sage_diagonal_matrix( J )
    L = sage_diagonal_matrix( [ d.abs().sqrt() for d in D.diagonal()] )
    U = L * W

    #
    # Do some tests
    #
    assert M_pol in ring.ideal( imp_lst )
    assert vx * M * vx == M_pol
    assert M * V == V * D

    #
    # output values
    #
    LSTools.p( 'quadratic form of signature (1,6) in ideal =', M_pol )
    LSTools.p( 'matrix M associated to quadratic form      =', list( M ) )
    LSTools.p( 'M == U.T*J*U                               =', list( U.T * J * U ) )
    LSTools.p( 'U                                          =', list( U ) )
    LSTools.p( 'J                                          =', list( J ) )


def usecase__get_base_points__and__get_linear_series():
    '''
    We obtain (infinitely near) base points of a 
    linear series defined over QQ. The base points
    are defined over a number field.
    '''
    ring = PolyRing( 'x,y,z', True )
    ls = LinearSeries( ['x^2+y^2', 'y^2+x*z'], ring )
    bp_tree = ls.get_bp_tree()
    LSTools.p( bp_tree )

    ls = LinearSeries.get( [ 2 ], bp_tree )
    LSTools.p( ls )
    LSTools.p( ls.get_bp_tree() )


def usecase__linear_normalization__and__adjoint():
    '''
    In this usecase are some applications of the 
    "linear_series" library.  we show by example 
    how to compute a linear normalization X of a surface Y, 
    or equivalently, how to compute the completion of a 
    linear series. Also we contruct an example of a surface Z 
    such that X is its adjoint surface.     
    '''
    #
    # Linear series corresponding to the parametrization
    # of a cubic surface X in P^4 that is the projection
    # of the Veronese embedding of P^2 into P^5.
    #
    ring = PolyRing( 'x,y,z', True )
    bp_tree = BasePointTree()
    bp = bp_tree.add( 'z', ( 0, 0 ), 1 )
    ls = LinearSeries.get( [2], bp_tree )
    imp_lst = ls.get_implicit_image()
    LSTools.p( 'linear series      =', ls.get_bp_tree() )
    LSTools.p( 'implicit image     =', imp_lst )

    #
    # compute Hilbert polynomial of X in QQ[x0,...,x6]
    #
    R = sage_PolynomialRing( sage_QQ, [ 'x' + str( i ) for i in range( len( ls.pol_lst ) )] )
    x_lst = R.gens()
    imp_lst = sage__eval( str( imp_lst ), R.gens_dict() )
    hpol = R.ideal( imp_lst ).hilbert_polynomial()
    hdeg = hpol.diff().diff()
    LSTools.p( 'Hilbert polynomial =', hpol )
    LSTools.p( 'implicit degree    =', hdeg )

    #
    # projection of X in P^4 to Y in P^3, where Y is singular.
    #
    ls = LinearSeries( ['x^2-x*y', 'x*z', 'y^2', 'y*z'], PolyRing( 'x,y,z', True ) )
    bp_tree = ls.get_bp_tree()
    LSTools.p( 'basepoint tree of projection =', bp_tree )
    eqn = ls.get_implicit_image()
    assert len( eqn ) == 1
    eqn = sage_SR( eqn[0] )
    x0, x1, x2, x3 = sage_var( 'x0,x1,x2,x3' )
    LSTools.p( 'eqn       =', eqn )
    LSTools.p( 'D(eqn,x0) =', sage_diff( eqn, x0 ) )
    LSTools.p( 'D(eqn,x1) =', sage_diff( eqn, x1 ) )
    LSTools.p( 'D(eqn,x2) =', sage_diff( eqn, x2 ) )
    LSTools.p( 'D(eqn,x3) =', sage_diff( eqn, x3 ) )

    #
    # compute normalization X of Y
    #
    ls_norm = LinearSeries.get( [2], bp_tree )
    LSTools.p( 'normalization = ', ls_norm )
    LSTools.p( '              = ', ls_norm.get_implicit_image() )

    #
    # define pre-adjoint surface Z
    #
    ring = PolyRing( 'x,y,z', True )
    bp_tree = BasePointTree()
    bp_tree.add( 'z', ( 0, 0 ), 2 )
    bp_tree.add( 'z', ( 0, 1 ), 1 )
    ls = LinearSeries.get( [5], bp_tree )
    LSTools.p( ls.get_bp_tree() )


def usecase__neron_severi_lattice():
    '''
    Compute NS-lattice of a linear series and
    the dimension of complete linear with base points. 
    '''

    # Blowup of projective plane in 3 colinear points
    # and 2 infinitly near points. The image of the
    # map associated to the linear series is a quartic
    # del pezzo surface with 5 families of conics. Moreover
    # the surface contains 8 straight lines.
    #
    ring = PolyRing( 'x,y,z', True )
    p1 = ( -1, 0 )
    p2 = ( 0, 0 )
    p3 = ( 1, 0 )
    p4 = ( 0, 1 )
    p5 = ( 2, 0 )
    bp_tree = BasePointTree()
    bp_tree.add( 'z', p1, 1 )
    bp_tree.add( 'z', p2, 1 )
    bp_tree.add( 'z', p3, 1 )
    bp = bp_tree.add( 'z', p4, 1 )
    bp.add( 't', p5, 1 )
    ls = LinearSeries.get( [3], bp_tree )
    LSTools.p( 'ls     = ', ls )
    LSTools.p( ls.get_bp_tree(), '\n\n   ', ls.get_implicit_image() )

    # Detects that 3 base points lie on a line.
    #
    bp_tree = BasePointTree()
    bp_tree.add( 'z', p1, 1 )
    bp_tree.add( 'z', p2, 1 )
    bp_tree.add( 'z', p3, 1 )
    ls123 = LinearSeries.get( [1], bp_tree )
    LSTools.p( 'ls123  =', ls123 )
    assert ls123.pol_lst == ring.coerce( '[y]' )

    # example of infinitly near base points
    # that is colinear with another simple base point.
    #
    bp_tree = BasePointTree()
    bp_tree.add( 'z', p1, 1 )
    bp = bp_tree.add( 'z', p4, 1 )
    bp.add( 't', ( 1, 0 ), 1 )
    ls1 = LinearSeries.get( [1], bp_tree )
    LSTools.p( 'ls1    =', ls1 )
    assert ls1.pol_lst == ring.coerce( '[x-y+z]' )

    # Detects that an infinitly near base points
    # is not colinear with other point.
    #
    for p in [p1, p2, p3]:
        bp_tree = BasePointTree()
        bp = bp_tree.add( 'z', p4, 1 )
        bp.add( 't', p5, 1 )
        bp_tree.add( 'z', p, 1 )
        ls45i = LinearSeries.get( [1], bp_tree )
        assert ls45i.pol_lst == []

    # line with predefined tangent
    #
    bp_tree = BasePointTree()
    bp = bp_tree.add( 'z', p4, 1 )
    bp.add( 't', p5, 1 )
    ls45 = LinearSeries.get( [1], bp_tree )
    LSTools.p( 'ls45   =', ls45 )
    assert ls45.pol_lst == ring.coerce( '[x-2*y+2*z]' )

    # class of conics through 5 basepoints
    #
    bp_tree = BasePointTree()
    bp_tree.add( 'z', p1, 1 )
    bp_tree.add( 'z', p2, 1 )
    bp_tree.add( 'z', p3, 1 )
    bp = bp_tree.add( 'z', p4, 1 )
    bp.add( 't', p5, 1 )
    ls1234 = LinearSeries.get( [2], bp_tree )
    LSTools.p( 'ls1234 =', ls1234 )
    LSTools.p( '\t\t', sage_factor( ls1234.pol_lst[0] ) )
    assert ring.coerce( '(x - 2*y + 2*z, 1)' ) in ring.aux_gcd( ls1234.pol_lst )[0]
    assert ring.coerce( '(y, 1)' ) in ring.aux_gcd( ls1234.pol_lst )[0]

    # class of conics through 4 basepoints
    # has a fixed component
    #
    bp_tree = BasePointTree()
    bp_tree.add( 'z', p4, 1 )
    ls4 = LinearSeries.get( [1], bp_tree )
    LSTools.p( 'ls4    =', ls4 )

    bp_tree = BasePointTree()
    bp_tree.add( 'z', p1, 1 )
    bp_tree.add( 'z', p2, 1 )
    bp_tree.add( 'z', p3, 1 )
    bp_tree.add( 'z', p4, 1 )
    ls1234 = LinearSeries.get( [2], bp_tree )
    LSTools.p( 'ls1234 =', ls1234 )

    # return

    # Compose map associated to linear series "ls"
    # with the inverse stereographic projection "Pinv".
    #
    R = sage_PolynomialRing( sage_QQ, 'y0,y1,y2,y3,y4,x,y,z' )
    y0, y1, y2, y3, y4, x, y, z = R.gens()
    delta = y1 ** 2 + y2 ** 2 + y3 ** 2 + y4 ** 2
    Pinv = [ y0 ** 2 + delta, 2 * y0 * y1, 2 * y0 * y2, 2 * y0 * y3, 2 * y0 * y4, -y0 ** 2 + delta]
    H = sage__eval( str( ls.pol_lst ), R.gens_dict() )
    PinvH = [ elt.subs( {y0:H[0], y1:H[1], y2:H[2], y3:H[3], y4:H[4]} ) for elt in Pinv ]
    LSTools.p( 'Pinv        =', Pinv )
    LSTools.p( 'H           =', H )
    LSTools.p( 'Pinv o H    =', PinvH )

    # In order to compute the NS-lattice of the image
    # of "PinvH" we do a base point analysis.
    #
    ls_PinvH = LinearSeries( [str( elt ) for elt in PinvH], ring )
    LSTools.p( 'ls_PinvH    =', ls_PinvH )
    LSTools.p( '\t\t', ls_PinvH.get_implicit_image() )

    if False:  # takes a long time
        LSTools.p( ls_PinvH.get_bp_tree() )



if __name__ == '__main__':

    ################################################
    #                                              #
    # Configuration of LSTools for debug output    #
    #                                              #
    ################################################

    LSTools.start_timer()
    mod_lst = []
    mod_lst += ['__main__.py']
    # mod_lst += ['get_linear_series.py']
    # mod_lst += ['get_solution_set.py']
    LSTools.filter( mod_lst )  # output only from specified modules
    # LSTools.filter( None )  # uncomment for showing all debug output

    ################################################
    #                                              #
    # End of configuration of LSTools              #
    #                                              #
    ################################################



    ################################################
    #                                              #
    # (Un)comment usecases for this package below. #
    #                                              #
    ################################################

    usecase__get_base_points__P2()  # shows how to obtain base points of linear series in the projective plane P^2
    usecase__get_base_points__P1P1()  # shows how to obtain base points of linear series in P^1xP^1
    usecase__get_base_points__examples()  # several examples of linear series and their base point
    usecase__get_linear_series__P2()  # shows how to construct a linear series in P^2 from given base points
    usecase__get_linear_series__P1P1_DP6()  # shows how to construct a linear series in P^1xP^1 from given base points
    usecase__get_implicit__DP6()  # shows how to compute the ideal of the surface parametrized by the linear series
    usecase__get_base_points__and__get_linear_series()  # another easy example for getting base points and linear series
    usecase__linear_normalization__and__adjoint()  # shows how to compute linear normalization and adjoint of surface
    usecase__neron_severi_lattice()  # compute NS-lattice and dimension of classes

    ###############################################
    #                                             #
    # End of list of usecase methods.             #
    #                                             #
    ###############################################

    LSTools.stop_timer()
    print
    print( 'The End' )




'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 4, 2016
@author: Niels Lubbes

The polynomial ring is a object that extends PolynomialRing of Sage
in the sense that the base field can be extended by roots, by passing
the minimal polynomial of the root as argument.
'''

from class_ls_tools import LSTools

from sage_interface import sage_QQ
from sage_interface import sage__eval
from sage_interface import sage_PolynomialRing
from sage_interface import sage_diff
from sage_interface import sage_factor
from sage_interface import sage_gcd
from sage_interface import sage_expand
from sage_interface import sage_SR
from sage_interface import sage_NumberField
from sage_interface import sage_FractionField
from sage_interface import sage_proof


class PolyRing:
    '''
    Polynomial ring defined over an algebraic number field.
    The base field is a number field and is a static variable.
    Unless explicitly specified in the constructor, the number field
    will contain roots, that were previously added.     
    '''

    # Initially the number field is the rational numbers (static).
    #
    num_field = sage_QQ

    # List of roots in number field that are adjoined to QQ (static).
    # This list is used in "PolyRing.__str__()" for the human readable
    # output format. Its elements consist of 2-tuples:
    #     ( [name of root], [minimal polynomial for root] ).
    #
    root_lst = []

    def __init__( self, var_lst = 'x,y,z', reset_num_field = False ):
        '''
        Constructor.
        
        INPUT:
            - "var_lst"         -- A string of characters separated by comma's.
                                   The characters represent the generators of a 
                                   polynomial ring over "Poly_Ring.num_field".
            - "reset_num_field" -- If True then the base field is reset to QQ.
                                   Otherwise, the base field remembers previously
                                   added roots.
        '''

        # reset static num_field variable
        #
        if reset_num_field:
            PolyRing.reset_base_field()

        # Polynomial ring over an algebraic number field.
        #
        self.pol_ring = sage_PolynomialRing( PolyRing.num_field, var_lst )

        # Usage: sage__eval('<element in pol_ring>', ring_dct)
        #
        self.ring_dct = {}

        # update dictionary
        #
        self.ring_dct['t'] = sage_PolynomialRing( PolyRing.num_field, 't' ).gens()[0]
        self.__update_ring_dct()

        # Unlike in PARI/GP, class group computations in Sage do not by default
        # assume the Generalized Riemann Hypothesis. To do class groups computations
        # not provably correctly we set proof.number_field(False). It can easily take
        # 1000 times longer to do computations with proof (the default).
        #
        sage_proof.number_field( False )


    @staticmethod
    def reset_base_field():
        '''
        OUTPUT:
            - Resets base field of this polynomial ring to QQ.
        '''
        PolyRing.num_field = sage_QQ
        PolyRing.root_lst = []


    def __update_ring_dct( self ):
        '''
        This private method updates "PolyRing.num_field" 
        and "self.ring_dct" and should be called if attributes of 
        "self" are modified.
        '''
        PolyRing.num_field = self.get_num_field()
        if PolyRing.num_field != sage_QQ:  # do not add {'1':1}
            self.ring_dct.update( self.get_num_field().gens_dict() )
        self.ring_dct.update( self.pol_ring.gens_dict() )
        self.ring_dct['t'] = self.ring_dct['t'].parent().change_ring( self.get_num_field() ).gens()[0]


    def set_num_field( self, field, root, min_pol ):
        '''
        INPUT:
            - field   -- A Sage field object. For example QQ.
            - root    -- An element in field that has been adjoined to it 
                         most recently.
            - min_pol -- The minimal polynomial of root.
            
        OUTPUT:
            - Modifies "self.pol_ring" to be a PolynomialRing over the new
              base field.
        '''
        self.pol_ring = self.pol_ring.change_ring( field )
        PolyRing.num_field = self.get_num_field()
        PolyRing.root_lst += [ ( root, min_pol ) ]
        self.__update_ring_dct()


    def get_num_field( self ):
        return self.pol_ring.base_ring()


    def set_gens( self, new_gens ):
        '''
        INPUT:
            - "new_gens" -- set of generators of PolynomialRing.
        OUTPUT:
            - Modifies "self.pol_ring" to be a PolynomialRing with 
              generators "new_gens" over the same number field as before.
        '''
        self.pol_ring = sage_PolynomialRing( self.get_num_field(), new_gens )
        self.__update_ring_dct()


    def gens( self ):
        '''
        OUTPUT:
            - Returns generators of polynomial ring.
        '''
        return self.pol_ring.gens()


    def root_gens( self ):
        '''
        OUTPUT:
            - Returns roots of number field which were adjoined to QQ. 
        '''
        gen_lst = list( self.get_num_field().gens() )
        gen_lst.reverse()
        return gen_lst


    def coerce( self, elt ):
        '''
        INPUT:
            - "elt"  -- A string of a list of polynomials in "self.pol_ring". 
        OUTPUT:
            - Coerces the string "elt" to an element of "self.pol_ring".
        '''
        return sage__eval( str( elt ), self.ring_dct )


    def coerce_ff( self, elt ):
        '''
        Some symbolic methods in sage are not available in the polynomial 
        ring over a number field, but are available in the polynomial ring
        over a fraction field.
        
        INPUT:
            - "elt"  -- String of an symbolic expression 
                        with polynomials in "self.pol_ring".
        OUTPUT:
            - Elements in a polynomial ring R with generators "self.gens()".
              The ground field of R is the fraction field of a polynomial ring 
              whose generators corresponds to the roots in the 
              number field "PolyRing.num_field".
              For example R=FF[x,y,z] where FF=FractionField( QQ[a0,a1,a2,a3] ). 
        '''
        eval_dct = {}

        # construct fraction field FF
        FF = sage_QQ
        if PolyRing.num_field != sage_QQ:
            ngens = PolyRing.num_field.gens_dict().keys()  # (a0,a1,...)
            FF = sage_FractionField( sage_PolynomialRing( sage_QQ, ngens ) )
            eval_dct.update( FF.gens_dict() )

        # construct polynomial ring R over FF
        pgens = self.pol_ring.gens_dict().keys()
        R = sage_PolynomialRing( FF, pgens )
        eval_dct.update( R.gens_dict() )

        # coerce "elt" to R
        return sage__eval( str( elt ), eval_dct )


    def coerce_sr( self, elt ):
        '''
        Some symbolic methods in sage are not available in the 
        polynomial ring over a number field, but are available 
        in the "SymbolicRing" of Sage.
        
        INPUT:
            - "elt"  -- String of expression with 
                        polynomials in "self.pol_ring".
        OUTPUT:
            - Elements in the "SymbolicRing" of Sage. This can be seen
              as a polynomial ring over QQ where generators correspond to
              the roots in "PolyRing.num_field" and "self.gens()".
              For example: QQ[a0,a1,a2,x,y,z]. 
        '''
        # construct dictionary for "sage_eval"
        sym_dct = self.ring_dct.copy()
        for key in sym_dct.keys():
            sym_dct[key] = sage_SR( str( sym_dct[key] ) )  # SR = Symbolic Ring

        # coerce "elt" to symbolic ring
        return sage__eval( str( elt ), sym_dct )


    def diff( self, pol, gen, mult ):
        '''
        INPUT:
            - "pol"  -- A polynomial in "self.pol_ring".
            - "gen"  -- A generator of "self.pol_ring".
            - "mult" -- An integer.            
        OUTPUT:
            - Derivative of "pol" wrt. variable "gen", "mult" times.         
        '''
        # currently there is a bug in Sage for
        # computing the derivative of a polynomial
        # defined over a number field

        spol, sgen = self.coerce_sr( [pol, gen] )
        dpol = sage_diff( spol, sgen, mult )
        dpol = self.coerce( dpol )

        return dpol


    def quo( self, pol1, pol2 ):
        '''
        INPUT:
            - "pol1" -- A polynomial in "self.pol_ring".
            - "pol2" -- A polynomial in "self.pol_ring".            
        OUTPUT:
            - Polynomial quotient of "pol1" and "pol2" in "self.pol_ring".         
        '''
        # currently there is a bug in Sage for
        # computing the polynomial quotient over a number field.

        spol1, spol2 = self.coerce_ff( [pol1, pol2] )
        squo = spol1.quo_rem( spol2 )[0]
        quo = self.coerce( squo )
        # LSTools.p( ( pol1, pol2, squo, quo ) )
        return quo


    def resultant( self, pol1, pol2, var ):
        '''
        INPUT:
            - "pol1" -- A polynomial in "self.pol_ring".
            - "pol2" -- A polynomial in "self.pol_ring".
            - "var"  -- A generator of "self.pol_ring".
        OUTPUT:
            - Resultant of "pol1" and "pol2" w.r.t. "var" in "self.pol_ring". 
        '''
        # currently there is a bug in Sage for computing
        # the resultant over a number field. So we coerce
        # the arguments first to a fraction field, then
        # we compute the resultant, after which we coerce
        # the resultant to be an element in PolyRing.

        spol1, spol2, svar = self.coerce_ff( [pol1, pol2, var] )

        if spol1 in sage_QQ or spol2 in sage_QQ:
            sres = -1
        else:
            sres = spol1.resultant( spol2, svar )

        res = self.coerce( sres )

        return res


    def aux_gcd( self, pol_lst ):
        '''
        INPUT:
            - "pol_lst" -- A (string of a) list of polynomials in "self.pol_ring"
        
        OUTPUT: 
            - Returns the following two lists:
                * A list of non-constant factors of the 
                  gcd over QQ of the polynomials:  
                    [( factor in "self.pol_ring", multiplicity ),...].
                * A list of polynomials in "self.pol_ring" such 
                  that the gcd is factored out.
        '''

        # In Sage it is currently not possible to compute gcd and to
        # factor a multivariate polynomial over a number field.

        # convert to symbolic ring
        spol_lst = self.coerce_sr( pol_lst )

        # compute gcd and factor in symbolic ring
        sgcd = sage_gcd( spol_lst )
        sfct_lst = sgcd.factor_list()

        # factor out gcd in polynomials
        spol_lst = [ sage_expand( spol / sgcd ) for spol in spol_lst]

        # coerce back
        fct_lst, npol_lst = self.coerce( [sfct_lst, spol_lst] )

        # ignore the factors that are constants
        nfct_lst = []
        for fct in fct_lst:
            if fct[0] not in PolyRing.num_field:
                nfct_lst += [ fct ]

        return nfct_lst, npol_lst


    def factor( self, pol ):
        '''
        INPUT:
            - "pol" -- A univariate polynomial in "self.pol_ring".
        OUTPUT
            - A list of factors of "pol" in "PolyRing.pol_ring":
              
              [  ( <polynomial-factor>, <multiplicity> ), ... ]               
              
              If "pol" is a constant then the empty list is returned.
        '''
        LSTools.p( 'factoring ', pol, ' in ', self )
        if pol in self.get_num_field():
            return []
        else:
            fct_lst = sage_factor( pol )
            LSTools.p( '\t', fct_lst )
            return fct_lst

        #
        # The call "sage_factor(pol)" sometimes raises the following warning:
        # "
        # local/lib/python2.7/site-packages/sage/rings/number_field/number_field.py:1526:
        # UserWarning: interpreting PARI polynomial 1 relative to the
        # defining polynomial x^2 + 1 of the PARI number field % (x, self.pari_polynomial()))
        # "
        #
        # If "pol" in "self.ring.get_num_field()" then "sage_factor( ypol )" raises
        # "
        # ArithmeticError: non-principal ideal in factorization
        # "
        #


    def ext_num_field( self, pol ):
        '''
        INPUT: 
            - "pol" --A (string of a) univariate polynomial over 
              "self.get_num_field()" in 't'.        
        METHOD:
            - Adjoins roots of "pol" to the number field of "self.ring". 
            - Returns self
        '''

        # factor
        pol = self.coerce( pol )
        if pol in self.get_num_field():
            return  # nothing to do here
        fct_lst = self.factor( pol )
        first_pol = fct_lst[0][0]


        if first_pol.degree() == 1:
            fct_lst = fct_lst[1:]
        else:

            # obtain extended number field with first factor
            gen_str = 'a' + str( len( self.root_lst ) )
            NF = sage_NumberField( [first_pol], gen_str )

            # set the new number field
            self.set_num_field( NF, NF.gens()[0], first_pol )


        # continue recursively with remaining factors until all factors are linear
        for fct in fct_lst:
            self.ext_num_field( fct[0] )

        return self


    def __str__( self ):
        '''
        OUTPUT:
          - A human readable string representation of object.
            The roots of the number field, that are adjoined 
            to the rational numbers QQ are depicted with their
            minimal polynomial.
            
        EXAMPLES:
          - sage: ring = PolyRing( 'x,y,z', True )
            sage: ring.ext_num_field( 't^2 + 1' )
            sage: ring.ext_num_field( 't^3 + a0' )
            sage: print( ring )
            out : QQ( <a0|t^2 + 1>, <a1|t^2 + a0*t - 1> )[x, y, z]
        '''
        s = 'QQ'
        if len( PolyRing.root_lst ) > 0:
            rs = ''
            for root in PolyRing.root_lst:
                rs += '<' + str( root[0] ) + '|' + str( root[1] ) + '>, '
            s += '( ' + rs[:-2] + ' )'
        s += str( self.gens() ).replace( '(', '[' ).replace( ')', ']' )
        return s






'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 4, 2016
@author: Niels Lubbes

This file declares 2 classes: "BasePointTree" and "BasePoint".
'''

class BasePointTree():
    '''
    This class represents a tree of base points of a linear series of 
    curves.
    '''

    # If True than the string representation of this object is short.
    short = True

    def __init__( self, chart_lst = ['z', 'x', 'y'] ):
        '''
        Constructor.
        '''

        # Linear series whose base points
        # are represented by this object.
        #
        self.ls = None

        # A list of charts where a chart is denoted by a String.
        # See documentation of "get_base_point_tree.in_previous_chart()"
        # for specification of chart strings.
        #
        self.chart_lst = chart_lst

        # Dictionary where a key is a chart
        # and a value is a list of BasePoint objects.
        # "self.chart_lst" is an ordered list of keys of this
        # dictionary.
        #
        self.chart_tree_dct = {}


    def add( self, chart, sol, mult ):
        '''
        INPUT:
            - "self"  -- BasePointTree object.
            - "chart" -- A String denoting a chart.
            - "sol"   -- A 2-tuple of elements in "PolyRing.num_field" 
                         representing a base point in a chart.     
            - "mult"  -- An integer representing the multiplicity of the base point.        
        OUTPUT:
            - Adds a base point to "self[chart]"
            - Return the added base point.
        '''

        bp = BasePoint( 0, chart, None )
        bp.sol = sol
        bp.mult = mult
        self[chart] += [bp]

        return bp


    # operator overloading for []
    def __getitem__( self, chart ):
        if chart not in self.chart_lst:
            raise ValueError( 'The chart key should be an element in:', self.chart_lst )

        if chart not in self.chart_tree_dct:
            self.chart_tree_dct[chart] = []

        return self.chart_tree_dct[chart]


    # operator overloading for []
    def __setitem__( self, chart, item ):
        if chart not in self.chart_lst:
            raise ValueError( 'The chart key should be an element in:', self.chart_lst )

        self.chart_tree_dct[chart] = item


    # overloads str(): human readable string representation of object
    def __str__( self ):
        '''                
        OUTPUT:
          - Human readable string representation of object. 
            The String consists of lines with the following format
                
                chart=[C], depth=[Integer], mult=[Integer], sol=[P], [LinearSeries]
 
            where 
                * C = An element of "self.chart_lst".
                * P = A 2-tuple: ( [PolyRing.num_field], [PolyRing.num_field] )
            and where 
                * sol  : A point P in the zeroset of the linear series.
                * chart: The current chart of the the point P.
                * mult : The multiplicity of P as a root.
                * depth: The depth when considering P as an infinitely near point
                         in a tree structure.
                * For each blowup chart we also depict the corresponding [LinearSeries]. 
            
            Note that the lines represent a tree structure.                     
            Below we see an example.
                
        EXAMPLE:
          - sage: ls = LinearSeries( ['x^2+y^2', 'x*z+y^2'], PolyRing( 'x,y,z', True ) )
            sage: print( ls.get_bp_tree() )  
            out :            
                { 2, <<x^2 + y^2, y^2 + x*z>>, QQ( <a0|t^2 + 1> )[x, y, z] }
                chart=z, depth=0, mult=1, sol=(0, 0), { 2, <<x^2 + y^2, y^2 + x>>, QQ( <a0|t^2 + 1> )[x, y] }
                    chart=t, depth=1, mult=1, sol=(0, 0), { 2, <<x^2*y + y, x + y>>, QQ( <a0|t^2 + 1> )[x, y] }
                chart=z, depth=0, mult=1, sol=(1, (-a0)), { 2, <<x^2 + y^2, y^2 + x>>, QQ( <a0|t^2 + 1> )[x, y] }
                chart=z, depth=0, mult=1, sol=(1, (a0)), { 2, <<x^2 + y^2, y^2 + x>>, QQ( <a0|t^2 + 1> )[x, y] }                                                      
        '''

        tree_str = ''
        if self.ls != None:
            tree_str += '\n' + str( self.ls )
        for chart in self.chart_lst:
            for bp in self[chart]:
                tree_str += str( bp )
        return tree_str.replace( '\n', '\n\t' )


    def alt_str( self ):
        ''' 
        This method can be useful for testing.
        
        OUTPUT:
          - A string representation without the linear series. 
        '''
        tree_str = ''
        for chart in self.chart_lst:
            for bp in self[chart]:
                tree_str += bp.alt_str()
        return tree_str.replace( '\n', '\n\t' )


class BasePoint():
    '''
    This class represents a binary tree of base points.
    If a base point has an infinitely near base point then 
    its 2 leaves represent two charts 's' and 't' of the blowup at these 
    base points.
    '''

    def __init__( self , depth, chart, ls ):
        '''
        Constructor.
        '''

        # Depth of base point tree.
        #
        self.depth = int( depth )

        # Chart denoted by a string.
        # See docs of "get_base_point_tree.in_previous_chart()"
        #
        self.chart = chart

        # LinearSeries
        #
        self.ls = ls

        # Base point represented as a 2-tuple of
        # elements in a number field.
        #
        self.sol = None

        # Multiplicity of a solution
        # (0=no solution, -1=overlapping chart)
        #
        self.mult = 0

        # lists of base points
        #
        self.bp_lst_t = []
        self.bp_lst_s = []


    def add( self, chart_st, sol, mult ):
        '''
        INPUT:
            - "self"     --
            - "chart_st" -- 's' or 't'
            - "sol"      -- A 2-tuple of elements in "PolyRing.num_field" 
                            representing an infinitely near base point in a blowup chart. 
            - "mult"     -- An integer representing the multiplicity of the base point. 
        OUTPUT:
            - Adds a base point to either "self.bp_lst_s" or "self.bp_lst_t".
            - Return the added infinitely near base point.
        '''

        if chart_st not in ['s', 't']:
            raise ValueError( 'Expecting "chart_st" to be either "s" or "t":', chart_st )

        bp = BasePoint( self.depth + 1, chart_st, None )
        bp.sol = sol
        bp.mult = mult

        dct = {'s':self.bp_lst_s, 't':self.bp_lst_t}
        dct[chart_st] += [bp]

        return bp

    def __str__( self ):
        '''
        OUTPUT:
          - Human readable string representation of object.
            See "BasePointTree.__str__()".
        '''

        if BasePointTree.short == True and self.mult in [0, -1]:
            return ''

        bp_str = ''
        bp_str += '\n' + 4 * self.depth * ' ' + 'chart=' + self.chart + ', '
        if self.mult == -1:
            bp_str += '(overlapping chart)' + ', '
        if self.mult == 0:
            bp_str += '(no solution)' + ', '
        bp_str += 'depth=' + str( self.depth ) + ', '
        bp_str += 'mult=' + str( self.mult ) + ', '
        bp_str += 'sol=' + str( self.sol ) + ', '
        bp_str += str( self.ls )

        for bp in self.bp_lst_t:
            bp_str += str( bp )

        for bp in self.bp_lst_s:
            bp_str += str( bp )

        return bp_str


    def alt_str( self ):
        ''' 
        This method can be useful for testing.
        
        OUTPUT:
          - A string representation without the linear series. 
        '''

        if self.mult in [0, -1]:
            return ''

        bp_str = ''
        bp_str += '\n' + 4 * self.depth * ' '
        bp_str += 'chart=' + self.chart + ', '
        bp_str += 'depth=' + str( self.depth ) + ', '
        bp_str += 'mult=' + str( self.mult ) + ', '
        bp_str += 'sol=' + str( self.sol ) + ', '

        for bp in self.bp_lst_t:
            bp_str += bp.alt_str()

        for bp in self.bp_lst_s:
            bp_str += bp.alt_str()

        return bp_str




'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 4, 2016
@author: Niels Lubbes

The method "get_bp_tree()" should be called from "LinearSeries.get_bp_tree()"
'''

from class_ls_tools import LSTools
from class_poly_ring import PolyRing
from class_base_points import BasePointTree
from class_base_points import BasePoint


def in_previous_chart( sol, chart ):
    '''
    We consider the following charts.
    
    * P^2(x:y:z) (projective plane). 
      There are 3 charts: x!=0, y!=0 and z!=0, 
      denoted by 'x', 'y' and 'z' resp.
      
    * P^1xP^1(x:y; v:w) (fiber product of projective line with itself).
      There are 4 charts: (x,v)!=(0,0), (x,w)!=(0,0), (y,v)!=(0,0) and (y,w)!=0,
      denoted by 'xv', 'xw', 'yv' and 'yw' respectively.      
    
    * A^2xP^1 (u,v; s:t) (fiber product of affine plane with projective line)
      There are 2 charts: s!=0 and t!=0, denoted by 's' and 't' resp.
      
    
    INPUT:
        - "sol"   -- Solution represented as a 2-tuple of elements 
                     in number field "PolyRing.num_field".
        - "chart" -- Chart over which solution is considered.
        
    OUTPUT:
        - Returns False if solution is already analyzed in a previous chart 
          and True otherwise.
    '''

    # sol = (a,b,mult)
    # Us --> Ut: (a,b)|-->(1/b,a*b)
    # Ut --> Us: (a,b)|-->(a*b,1/a)

    if chart == 't':
        return False
    elif chart == 's':
        return sol[1] != 0
    elif chart == 'z':  # sol=(x,y)
        return False
    elif chart == 'y':  # sol=(x,z)
        return sol[1] != 0
    elif chart == 'x':  # sol=(y,z)
        return sol[0] != 0 or sol[1] != 0
    elif chart == 'xv':  # sol=(y,w)
        return False
    elif chart == 'xw':  # sol=(y,v)
        return sol[1] != 0
    elif chart == 'yv':  # sol=(x,w)
        return sol[0] != 0
    elif chart == 'yw':  # sol=(x,v)
        return sol[0] != 0 or sol[1] != 0
    else:
        raise ValueError( 'chart =' + chart )


def get_bp_lst_chart( ls, chart, depth = 0 ):
    '''
    INPUT:
        - "ls"    -- LinearSeries.
        - "chart" -- A String representing the current chart 
                     of linear series "ls".  
        - "depth" -- Integer representing the recursive depth 
                     of this function call.
    OUTPUT:
        - Returns a list of BasePoint objects representing base points
          of linear series "ls".
    '''

    LSTools.p( 'input =', ( depth, chart, str( ls ) ) )

    nls = ls.copy()
    if chart == 's':
        nls.pol_lst = [ nls.ring.gens()[0] ] + nls.pol_lst
    elif chart == 't':
        nls.pol_lst = [ nls.ring.gens()[1] ] + nls.pol_lst

    sol_lst = nls.get_solution_set()

    bp_lst = []

    # if no solutions, add empty base point with BasePoint.mult=0.
    if sol_lst == []:
        bp_lst += [ BasePoint( depth, chart, ls ) ]

    for sol in sol_lst:

        bp = BasePoint( depth, chart, ls )
        bp.sol = sol

        if in_previous_chart( sol, chart ):

            bp.mult = -1  # this indicates that base point was already considered.

        else:

            ls_t, bp.mult = ls.copy().translate_to_origin( sol ).blow_up_origin( 't' )
            bp.bp_lst_t = get_bp_lst_chart( ls_t, 't', depth + 1 )

            ls_s, mult_s = ls.copy().translate_to_origin( sol ).blow_up_origin( 's' )
            bp.bp_lst_s = get_bp_lst_chart( ls_s, 's', depth + 1 )

            if bp.mult != mult_s:
                raise Exception( 'Unexpected multiplicities: (bp.mult, mult_s) =', ( bp.mult, mult_s ) )

        bp_lst += [ bp ]

    return bp_lst


def get_bp_tree( ls ):
    '''
    INPUT:
        - "ls" -- A LinearSeries. Its polynomials are 
                  either 
                      homogenous in (x:y:z) 
                  or
                      bi-homogenous in (x:y) and (v:w).
    OUPUT:
        - A BasePointTree representing the simple base points 
          and infinitely near base points of "ls".
    '''

    # create BasePointTree object
    g_set = set( [str( g ) for g in ls.gens()] )
    if g_set == set( ['x', 'y', 'v', 'w'] ):
        bp_tree = BasePointTree( ['xv', 'xw', 'yv', 'yw'] )
    elif g_set == set( ['x', 'y', 'z'] ):
        bp_tree = BasePointTree( ['z', 'x', 'y'] )
    else:
        ValueError( 'Expect "ls.gens()" to be a subset of:', ['z', 'x', 'y', 'v', 'w'] )
    bp_tree.ls = ls

    # get base points for each affine plane chart of linear series
    for c in bp_tree.chart_lst:
        ls_c = ls.copy().chart( c )
        bp_tree[c] = get_bp_lst_chart( ls_c, c )

    return bp_tree





